//------------------------------------------------------------------------------
// <copyright file="ConnectionManagerChannel.h" company="LiquidRobotics">
//
// Copyright (c) 2014 Liquid Robotics, Inc.
//
// Permission is hereby granted, free of charge, to any person obtaining a copy
// of this software and associated documentation files (the "Software"), to deal
// in the Software without restriction, including without limitation the rights
// to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
// copies of the Software, and to permit persons to whom the Software is
// furnished to do so, subject to the following conditions:
//
// The above copyright notice and this permission notice shall be included in
// all copies or substantial portions of the Software.
//
// THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
// IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
// FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
// AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
// LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
// OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
// THE SOFTWARE.
//
// This software is associated with a release.  Liquid Robotics makes no
// representation that the design, interfaces or component distribution will
// remain unchanged in future releases.
//
// </copyright>
//
// <summary>
// 	Declaration of ConnectionManagerChannel class, a serializable entity that
//  describes a Channel, or a Communications transport mechanism
// </summary>
//
// <owner>Arash Ushani</owner>
//------------------------------------------------------------------------------

#ifndef __ConnectionManagerChannel_H__
#define __ConnectionManagerChannel_H__

#include "ConnectionManagerChannelSettings.h"
#include "LqrSMACLogger.h"
#include "LqrSMACSharedLib.h"
#include <string>
#include <boost/signals2.hpp>
#include <boost/thread.hpp>

namespace LiquidRobotics {
namespace ConnectionManager {

class SHARED_LIB_PUBLIC ConnectionManagerChannel
{
public:

	typedef boost::signals2::scoped_connection ScopedConnection;
	typedef boost::shared_ptr<ScopedConnection> ScopedConnectionPtr;
	typedef boost::signals2::signal<void (ConnectionManagerChannel& sender, const int& signalStrength)> SignalStrengthSignalType;
	typedef boost::shared_ptr<SignalStrengthSignalType> SignalStrengthSignalTypePtr;
	typedef boost::signals2::mutex Mutex;
	typedef boost::shared_ptr<Mutex> MutexPtr;
	typedef std::string string;
	typedef LiquidRobotics::SMACFramework::Logger Logger;

	ConnectionManagerChannel(const ConnectionManagerChannelSettings& _settings, Logger::Ptr _logger);
	ConnectionManagerChannel(const ConnectionManagerChannel& rhs);

	int GetProtocolRevision();

	virtual bool ChannelOn();
	virtual bool ChannelOff();

	void lock();
	bool try_lock();
	void unlock();

	ScopedConnectionPtr SubscribeToSignalStrength(SignalStrengthSignalType::slot_type slot);

	bool ReportsSignalStrength();

	std::string Describe();

//	const ConnectionManagerChannelSettings& GetChannelSettings();

	int GetSignalStrength();


protected:

	void SignalStrengthThread();

	MutexPtr channelMutex;

	SignalStrengthSignalTypePtr signalStrengthSignalPtr;

	ConnectionManagerChannelSettings settings;
	Logger::Ptr logger;
	boost::thread thread;
};


}
}

BOOST_CLASS_VERSION(LiquidRobotics::ConnectionManager::ConnectionManagerChannel, 1)

#endif
